/*--------------------------------*-C-*---------------------------------*
 * File:	main.c
 *----------------------------------------------------------------------*
 * $Id: main.c,v 1.131 2000/09/19 06:31:17 gcw Exp $
 *
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 1992      John Bovey, University of Kent at Canterbury <jdb@ukc.ac.uk>
 *				- original version
 * Copyright (C) 1994      Robert Nation <nation@rocket.sanders.lockheed.com>
 *				- extensive modifications
 * Copyright (C) 1995      Garrett D'Amore <garrett@netcom.com>
 * Copyright (C) 1997      mj olesen <olesen@me.QueensU.CA>
 *				- extensive modifications
 * Copyright (C) 1997,1998 Oezguer Kesim <kesim@math.fu-berlin.de>
 * Copyright (C) 1998,1999,2000 Geoff Wing <gcw@pobox.com>
 *				- extensive modifications
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *---------------------------------------------------------------------*/

#include "../config.h"		/* NECESSARY */
#define  INTERN			/* assign all global vars to me */
#include "rxvt.h"		/* NECESSARY */
#include "main.intpro"		/* PROTOS for internal routines */

#include <signal.h>

#ifdef TTY_GID_SUPPORT
# include <grp.h>
#endif

#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

/*----------------------------------------------------------------------*/
/* rxvt_init() */
/* EXTPROTO */
int
rxvt_init(rxvt_t *r, int argc, const char *const *argv)
{
    const char    **cmd_argv;

    _rxvt_vars = r;		/* only assignment to _rxvt_vars */
    if (rxvt_init_vars(r) < 0)
	exit(EXIT_FAILURE);

/*
 * Save and then give up any super-user privileges
 * If we need privileges in any area then we must specifically request it.
 * We should only need to be root in these cases:
 *  1.  write utmp entries on some systems
 *  2.  chown tty on some systems
 */
    rxvt_privileges(r, SAVE);
    rxvt_privileges(r, IGNORE);

    rxvt_init_secondary(r);

    cmd_argv = rxvt_init_resources(r, argc, argv);

#if (MENUBAR_MAX)
    rxvt_menubar_read(r, r->h->rs[Rs_menu]);
#endif
    rxvt_scrollbar_mapping(r, r->Options & Opt_scrollBar);

    rxvt_Create_Windows(r, argc, argv);

    rxvt_init_xlocale(r);

    rxvt_scr_reset(r);		/* initialize screen */
    rxvt_Gr_reset(r);		/* reset graphics */

#ifdef DEBUG_X
    XSynchronize(r->Xdisplay, True);
    XSetErrorHandler((XErrorHandler) abort);
#else
    XSetErrorHandler((XErrorHandler) rxvt_xerror_handler);
#endif

    if (scrollbar_visible()) {
	rxvt_Resize_scrollBar(r);
	XMapWindow(r->Xdisplay, r->scrollBar.win);
    }
#if (MENUBAR_MAX)
    if (menubar_visible())
	XMapWindow(r->Xdisplay, r->menuBar.win);
#endif
#ifdef TRANSPARENT
    if (r->Options & Opt_transparent) {
	XSelectInput(r->Xdisplay, Xroot, PropertyChangeMask);
	rxvt_check_our_parents(r);
    }
#endif
    XMapWindow(r->Xdisplay, r->TermWin.vt);
    XMapWindow(r->Xdisplay, r->TermWin.parent[0]);

    rxvt_init_env(r);
    rxvt_init_command(r, cmd_argv);
    return 0;
}

/* ------------------------------------------------------------------------- *
 *                       SIGNAL HANDLING & EXIT HANDLER                      *
 * ------------------------------------------------------------------------- */
/*
 * Catch a SIGCHLD signal and exit if the direct child has died
 */
/* ARGSUSED */
/* EXTPROTO */
RETSIGTYPE
rxvt_Child_signal(int unused)
{
    int             pid, save_errno = errno;

    do {
	errno = 0;
    } while ((pid = waitpid(-1, NULL, WNOHANG)) == -1 && errno == EINTR);

    if (pid == _rxvt_vars->h->cmd_pid)
	exit(EXIT_SUCCESS);

    errno = save_errno;
    signal(SIGCHLD, rxvt_Child_signal);
}

/*
 * Catch a fatal signal and tidy up before quitting
 */
/* EXTPROTO */
RETSIGTYPE
rxvt_Exit_signal(int sig)
{
    signal(sig, SIG_DFL);
#ifdef DEBUG_CMD
    rxvt_print_error("signal %d", sig);
#endif
    rxvt_clean_exit();
    kill(getpid(), sig);
}

/* ARGSUSED */
/* INTPROTO */
int
rxvt_xerror_handler(const Display *display, const XErrorEvent *event)
{
    rxvt_t         *r = _rxvt_vars;

    if (r->h->allowedxerror == -1) {
	r->h->allowedxerror = event->error_code;
	return 0;		/* ignored anyway */
    }
    rxvt_print_error("XError: Request: %d . %d, Error: %d",
		     event->request_code, event->minor_code,
		     event->error_code);
    /* XXX: probably should call rxvt_clean_exit() bypassing X routines */
    exit(EXIT_FAILURE);
    /* NOTREACHED */
}

/*----------------------------------------------------------------------*/
/*
 * Exit gracefully, clearing the utmp entry and restoring tty attributes
 * TODO: if debugging, this should free up any known resources if we can
 */
/* EXTPROTO */
void
rxvt_clean_exit(void)
{
    rxvt_t         *r = _rxvt_vars;

#ifdef DEBUG_SCREEN
    rxvt_scr_release(r);
#endif
    rxvt_privileged_ttydev(r, RESTORE);
    rxvt_privileged_utmp(r, RESTORE);
#ifdef USE_XIM
    if (r->h->Input_Context != NULL) {
	XDestroyIC(r->h->Input_Context);
	r->h->Input_Context = NULL;
    }
#endif
}

/* ------------------------------------------------------------------------- *
 *                            PRIVILEGED OPERATIONS                          *
 * ------------------------------------------------------------------------- */
/* take care of suid/sgid super-user (root) privileges */
/* INTPROTO */
void
rxvt_privileges(rxvt_t *r, int mode)
{
#if ! defined(__CYGWIN32__)
# if !defined(HAVE_SETEUID) && defined(HAVE_SETREUID)
/* setreuid() is the poor man's setuid(), seteuid() */
#  define seteuid(a)	setreuid(-1, (a))
#  define setegid(a)	setregid(-1, (a))
#  define HAVE_SETEUID
# endif
# ifdef HAVE_SETEUID
    switch (mode) {
    case IGNORE:
	/*
	 * change effective uid/gid - not real uid/gid - so we can switch
	 * back to root later, as required
	 */
	seteuid(getuid());
	setegid(getgid());
	break;
    case SAVE:
	r->h->euid = geteuid();
	r->h->egid = getegid();
	break;
    case RESTORE:
	seteuid(r->h->euid);
	setegid(r->h->egid);
	break;
    }
# else
    switch (mode) {
    case IGNORE:
	setuid(getuid());
	setgid(getgid());
    /* FALLTHROUGH */
    case SAVE:
    /* FALLTHROUGH */
    case RESTORE:
	break;
    }
# endif
#endif
}

#ifdef UTMP_SUPPORT
/* EXTPROTO */
void
rxvt_privileged_utmp(rxvt_t *r, char action)
{
    D_MAIN((stderr, "rxvt_privileged_utmp(%c); waiting for: %c (pid: %d)", action, r->h->next_utmp_action, getpid()));
    if (r->h->next_utmp_action != action || action == IGNORE)
	return;
    if (r->h->ttydev == NULL || *r->h->ttydev == '\0')
	return;
    if (!(r->Options & Opt_utmpInhibit)) {
	rxvt_privileges(r, RESTORE);
	if (action == SAVE) {
	    r->h->next_utmp_action = RESTORE;
	    rxvt_makeutent(r, r->h->ttydev, r->h->rs[Rs_display_name]);
	} else if (action == RESTORE) {
	    r->h->next_utmp_action = IGNORE;
	    rxvt_cleanutent(r);
	}
	rxvt_privileges(r, IGNORE);
    }
}
#endif

#ifndef __CYGWIN32__
/* EXTPROTO */
void
rxvt_privileged_ttydev(rxvt_t *r, char action)
{
    D_MAIN((stderr, "rxvt_privileged_ttydev(r, %c); waiting for: %c (pid: %d)", action, r->h->next_tty_action, getpid()));
    if (r->h->next_tty_action != action || action == IGNORE
	|| !r->h->changettyowner)
	return;
    if (r->h->ttydev == NULL || *r->h->ttydev == '\0')
	return;

    rxvt_privileges(r, RESTORE);

    if (action == RESTORE) {
	r->h->next_tty_action = IGNORE;
# ifndef RESET_TTY_TO_COMMON_DEFAULTS
	chmod(r->h->ttydev, r->h->ttyfd_stat.st_mode);
	chown(r->h->ttydev, r->h->ttyfd_stat.st_uid, r->h->ttyfd_stat.st_gid);
# else
	chmod(r->h->ttydev, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH
			     | S_IWOTH));
	chown(r->h->ttydev, 0, 0);
# endif
    } else if (action == SAVE) {
	r->h->next_tty_action = RESTORE;

# ifndef RESET_TTY_TO_COMMON_DEFAULTS
/* store original tty status for restoration rxvt_clean_exit() -- rgg 04/12/95 */
	if (lstat(r->h->ttydev, &r->h->ttyfd_stat) < 0)	/* you lose out */
	    r->h->next_tty_action = IGNORE;
	else
# endif
	{
	    chown(r->h->ttydev, getuid(), r->h->ttygid); /* fail silently */
	    chmod(r->h->ttydev, r->h->ttymode);
# ifdef HAVE_REVOKE
	    revoke(r->h->ttydev);
# endif
	}
    }

    rxvt_privileges(r, IGNORE);

# ifndef RESET_TTY_TO_COMMON_DEFAULTS
    D_MAIN((stderr, "%s \"%s\": mode %03o, uid %d, gid %d", action == RESTORE ? "Restoring" : (action == SAVE ? "Saving" : "UNKNOWN ERROR for"), r->h->ttydev, r->h->ttyfd_stat.st_mode, r->h->ttyfd_stat.st_uid, r->h->ttyfd_stat.st_gid));
# endif
}
#endif

/*----------------------------------------------------------------------*/
/* EXTPROTO */
void
rxvt_szhints_set(rxvt_t *r)
{
    int             x, y, flags;
    unsigned int    width, height;

    r->szHint.flags = PMinSize | PResizeInc | PBaseSize | PWinGravity;
    r->szHint.win_gravity = NorthWestGravity;
    r->szHint.min_aspect.x = r->szHint.min_aspect.y = 1;

    flags = (r->h->rs[Rs_geometry] ? XParseGeometry(r->h->rs[Rs_geometry], &x,
						    &y, &width, &height)
				   : 0);

    if (flags & WidthValue) {
	r->TermWin.ncol = min(width, MAX_COLS);
	r->szHint.flags |= USSize;
    }
    if (flags & HeightValue) {
	r->TermWin.nrow = min(height, MAX_ROWS);
	r->szHint.flags |= USSize;
    }
    r->TermWin.width = r->TermWin.ncol * r->TermWin.fwidth;
    r->TermWin.height = r->TermWin.nrow * r->TermWin.fheight;
    rxvt_szhints_recalc(r);

    if (flags & XValue) {
	if (flags & XNegative) {
	    x += (DisplayWidth(r->Xdisplay, Xscreen) - r->szHint.width
		  - 2 * r->TermWin.ext_bwidth);
	    r->szHint.win_gravity = NorthEastGravity;
	}
	r->szHint.x = x;
	r->szHint.flags |= USPosition;
    }
    if (flags & YValue) {
	if (flags & YNegative) {
	    y += (DisplayHeight(r->Xdisplay, Xscreen) - r->szHint.height
		  - 2 * r->TermWin.ext_bwidth);
	    if (r->szHint.win_gravity == NorthEastGravity)
		r->szHint.win_gravity = SouthEastGravity;
	    else
		r->szHint.win_gravity = SouthWestGravity;
	}
	r->szHint.y = y;
	r->szHint.flags |= USPosition;
    }
}

/* INTPROTO */
void
rxvt_szhints_recalc(rxvt_t *r)
{
    r->szHint.base_width = 2 * r->TermWin.int_bwidth;
    r->szHint.base_height = 2 * r->TermWin.int_bwidth;
    r->szHint.base_width += (scrollbar_visible()
			     ? (r->scrollBar.width + 2 * r->sb_shadow)
			     : 0);
    r->szHint.base_height += (menubar_visible() ? menuBar_TotalHeight() : 0);
    r->szHint.width_inc = r->TermWin.fwidth;
    r->szHint.height_inc = r->TermWin.fheight;
    r->szHint.min_width = r->szHint.base_width + r->szHint.width_inc;
    r->szHint.min_height = r->szHint.base_height + r->szHint.height_inc;
    r->szHint.width = r->szHint.base_width + r->TermWin.width;
    r->szHint.height = r->szHint.base_height + r->TermWin.height;
}

/*----------------------------------------------------------------------*/
/*
 * Tell the teletype handler what size the window is.
 * Called after a window size change.
 */
/* EXTPROTO */
void
rxvt_tt_winsize(rxvt_t *r, int fd)
{
    struct winsize  ws;

    if (fd < 0)
	return;

    ws.ws_col = (unsigned short)r->TermWin.ncol;
    ws.ws_row = (unsigned short)r->TermWin.nrow;
    ws.ws_xpixel = ws.ws_ypixel = 0;
    ioctl(fd, TIOCSWINSZ, &ws);
}

/* EXTPROTO */
void
rxvt_tt_resize(rxvt_t *r)
{
    rxvt_tt_winsize(r, r->cmd_fd);
}

/*----------------------------------------------------------------------*/
/* rxvt_change_font() - Switch to a new font */
/*
 * init = 1   - initialize
 *
 * fontname == FONT_UP  - switch to bigger font
 * fontname == FONT_DN  - switch to smaller font
 */
/* EXTPROTO */
void
rxvt_change_font(rxvt_t *r, int init, const char *fontname)
{
    const char     *msg = "can't load font \"%s\"";
    int             fh, fw, recheckfonts;
    int             idx = 0;	/* index into r->h->rs[Rs_font] */
    XFontStruct    *xfont;
#ifdef MULTICHAR_SET
    int             i;
    char           *c, *enc;
    char            tmpbuf[64];
#endif

#if (FONT0_IDX == 0)
# define IDX2FNUM(i)	(i)
# define FNUM2IDX(f)	(f)
#else
# define IDX2FNUM(i)	((i) == 0 ? FONT0_IDX : ((i) <= FONT0_IDX ? ((i)-1) : (i)))
# define FNUM2IDX(f)	((f) == FONT0_IDX ? 0 : ((f) < FONT0_IDX ? ((f)+1) : (f)))
#endif
#define FNUM_RANGE(i)	((i) <= 0 ? 0 : ((i) >= NFONTS ? (NFONTS-1) : (i)))

    if (init) {
#ifndef NO_BOLDFONT
	r->TermWin.boldFont_loaded = NULL;
#endif
	r->h->fnum = FONT0_IDX;	/* logical font number */
    } else {
	switch (fontname[0]) {
	case '\0':
	    r->h->fnum = FONT0_IDX;
	    fontname = NULL;
	    break;

	    /* special (internal) prefix for font commands */
	case FONT_CMD:
	    idx = atoi(fontname + 1);
	    switch (fontname[1]) {
	    case '+':		/* corresponds to FONT_UP */
		r->h->fnum += (idx ? idx : 1);
		r->h->fnum = FNUM_RANGE(r->h->fnum);
		break;

	    case '-':		/* corresponds to FONT_DN */
		r->h->fnum += (idx ? idx : -1);
		r->h->fnum = FNUM_RANGE(r->h->fnum);
		break;

	    default:
		if (fontname[1] != '\0' && !isdigit(fontname[1]))
		    return;
		if (idx < 0 || idx >= (NFONTS))
		    return;
		r->h->fnum = IDX2FNUM(idx);
		break;
	    }
	    fontname = NULL;
	    break;

	default:
	    if (fontname != NULL) {
		/* search for existing fontname */
		for (idx = 0; idx < NFONTS; idx++) {
		    if (!STRCMP(r->h->rs[Rs_font + idx], fontname)) {
			r->h->fnum = IDX2FNUM(idx);
			fontname = NULL;
			break;
		    }
		}
	    } else
		return;
	    break;
	}
	/* re-position around the normal font */
	idx = FNUM2IDX(r->h->fnum);

	if (fontname != NULL) {
	    char           *name;

	    xfont = XLoadQueryFont(r->Xdisplay, fontname);
	    if (!xfont)
		return;

	    name = MALLOC(STRLEN(fontname + 1) * sizeof(char));

	    if (name == NULL) {
		XFreeFont(r->Xdisplay, xfont);
		return;
	    }
	    STRCPY(name, fontname);
	    if (r->h->newfont[idx] != NULL)
		FREE(r->h->newfont[idx]);
	    r->h->newfont[idx] = name;
	    r->h->rs[Rs_font + idx] = r->h->newfont[idx];
	}
    }
    if (r->TermWin.font)
	XFreeFont(r->Xdisplay, r->TermWin.font);

/* load font or substitute */
    xfont = XLoadQueryFont(r->Xdisplay, r->h->rs[Rs_font + idx]);
    if (!xfont) {
	rxvt_print_error(msg, r->h->rs[Rs_font + idx]);
	r->h->rs[Rs_font + idx] = "fixed";
	xfont = XLoadQueryFont(r->Xdisplay, r->h->rs[Rs_font + idx]);
	if (!xfont) {
	    rxvt_print_error(msg, r->h->rs[Rs_font + idx]);
	    goto Abort;
	}
    }
    r->TermWin.font = xfont;

#ifndef NO_BOLDFONT
/* fail silently */
    if (init && r->h->rs[Rs_boldFont] != NULL)
	r->TermWin.boldFont_loaded = XLoadQueryFont(r->Xdisplay,
						    r->h->rs[Rs_boldFont]);
#endif

/* alter existing GC */
    if (!init) {
	XSetFont(r->Xdisplay, r->TermWin.gc, r->TermWin.font->fid);
	rxvt_menubar_expose(r);
    }

/* set the sizes */
    fw = rxvt_get_fontwidest(r->TermWin.font);
    fh = r->TermWin.font->ascent + r->TermWin.font->descent;
    if (fw == r->TermWin.font->min_bounds.width)
	r->TermWin.fprop = 0;	/* Mono-spaced (fixed width) font */
    else
	r->TermWin.fprop = 1;	/* Proportional font */
    recheckfonts = !(fw == r->TermWin.fwidth && fh == r->TermWin.fheight);
    r->TermWin.fwidth = fw;
    r->TermWin.fheight = fh;

/* check that size of boldFont is okay */
#ifndef NO_BOLDFONT
    if (recheckfonts) {
	r->TermWin.boldFont = NULL;
	if (r->TermWin.boldFont_loaded != NULL) {
	    fw = rxvt_get_fontwidest(r->TermWin.boldFont_loaded);
	    fh = r->TermWin.boldFont_loaded->ascent
		 + r->TermWin.boldFont_loaded->descent;
	    if (fw <= r->TermWin.fwidth && fh <= r->TermWin.fheight)
		r->TermWin.boldFont = r->TermWin.boldFont_loaded;
	    r->TermWin.bprop =
		!(fw == r->TermWin.fwidth /* && fh == r->TermWin.fheight */ );
	}
    }
#endif				/* NO_BOLDFONT */

#ifdef MULTICHAR_SET
    if (r->TermWin.mfont)
	XFreeFont(r->Xdisplay, r->TermWin.mfont);

/* load font or substitute */
    if (r->h->rs[Rs_mfont + idx] == NULL
	|| (xfont =
	    XLoadQueryFont(r->Xdisplay, r->h->rs[Rs_mfont + idx])) == NULL) {
	/* TODO: improve this */
	i = 0;
	switch (r->encoding_method) {
	case GB:
	    c = "-*-r-*-%.2d-*-gb2312.1980-0";
	    enc = "GB";
	    break;
	case BIG5:
	    c = "-*-r-*-%.2d-*-big5-0";
	    enc = "BIG5";
	    break;
	case EUCJ:
	case SJIS:
	    c = "-*-%.2d-*-jisx0208*-*";
	    enc = "EUCJ/SJIS";
	    break;
	case EUCKR:
	    c = "-*-%.2d-*-ksc5601*-*";
	    enc = "KR";
	    break;
	default:
	    i = fh;		/* jump past next two sections */
	    break;
	}
	for (; i < fh / 2; i++) {
	    sprintf(tmpbuf, c, fh - i);
	    xfont = XLoadQueryFont(r->Xdisplay, tmpbuf);
	    if (xfont) {
		r->h->rs[Rs_mfont + idx] = MALLOC(STRLEN(tmpbuf) + 1);
		STRCPY(r->h->rs[Rs_mfont + idx], tmpbuf);
		break;
	    }
	}
	if (xfont == NULL && i != fh)
	    rxvt_print_error("no similar multichar font: encoding %s; size %d",
			     enc, fh);
    }
    r->TermWin.mfont = xfont;

    if (recheckfonts)
	/* XXX: This checks what? */
	if (r->TermWin.mfont != NULL) {
	    fw = rxvt_get_fontwidest(r->TermWin.mfont);
	    fh = r->TermWin.mfont->ascent + r->TermWin.mfont->descent;
	    if (fw > (r->TermWin.fwidth * 2) || fh > r->TermWin.fheight)
		r->TermWin.mfont = NULL;
	    r->TermWin.mprop = !(fw == r->TermWin.fwidth
				 /* && fh == r->TermWin.fheight */ );
	}
#endif				/* MULTICHAR_SET */

    /* Fontset setting is only valid when xlocale initialized. Since
     * rxvt_init_xlocale() is called after here, so we don't set fontset when
     * initialization, but let it set by rxvt_init_xlocale() */
    if (init)
	rxvt_setTermFontSet(r, -1);
    else
	rxvt_setTermFontSet(r, idx);

    rxvt_set_colorfgbg(r);

    r->TermWin.width = r->TermWin.ncol * r->TermWin.fwidth;
    r->TermWin.height = r->TermWin.nrow * r->TermWin.fheight;

    if (!init) {
	rxvt_resize_all_windows(r);
	rxvt_scr_touch(r, True);
    }
    r->h->num_scr_allow = 0;
    return;
  Abort:
    rxvt_print_error("aborting");	/* fatal problem */
    exit(EXIT_FAILURE);
#undef IDX2FNUM
#undef FNUM2IDX
#undef FNUM_RANGE
    /* NOTREACHED */
}

#ifdef STRICT_FONT_CHECKING
/* INTPROTO */
int
rxvt_get_fontwidest(XFontStruct *f)
{
    int             i, cw, fw = 0;

    if (f->min_bounds.width == f->max_bounds.width)
	return f->min_bounds.width;
    if (f->per_char == NULL)
	return f->max_bounds.width;
    for (i = f->max_char_or_byte2 - f->min_char_or_byte2; --i >= 0;) {
	cw = f->per_char[i].width;
	MAX_IT(fw, cw);
    }
    return fw;
}
#endif

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/* xterm sequences - title, iconName, color (exptl) */
/* EXTPROTO */
void
rxvt_set_title(rxvt_t *r, const char *str)
{
#ifndef SMART_WINDOW_TITLE
    XStoreName(r->Xdisplay, r->TermWin.parent[0], str);
#else
    char           *name;

    if (XFetchName(r->Xdisplay, r->TermWin.parent[0], &name) == 0)
	name = NULL;
    if (name == NULL || STRCMP(name, str))
	XStoreName(r->Xdisplay, r->TermWin.parent[0], str);
    if (name)
	XFree(name);
#endif
}

/* EXTPROTO */
void
rxvt_set_iconName(rxvt_t *r, const char *str)
{
#ifndef SMART_WINDOW_TITLE
    XSetIconName(r->Xdisplay, r->TermWin.parent[0], str);
#else
    char           *name;

    if (XGetIconName(r->Xdisplay, r->TermWin.parent[0], &name))
	name = NULL;
    if (name == NULL || STRCMP(name, str))
	XSetIconName(r->Xdisplay, r->TermWin.parent[0], str);
    if (name)
	XFree(name);
#endif
}

#ifdef XTERM_COLOR_CHANGE
/* EXTPROTO */
void
rxvt_set_window_color(rxvt_t *r, int idx, const char *color)
{
    XColor          xcol;
    int             i;

    if (color == NULL || *color == '\0')
	return;

/* handle color aliases */
    if (isdigit(*color)) {
	i = atoi(color);
	if (i >= 8 && i <= 15) {	/* bright colors */
	    i -= 8;
# ifndef NO_BRIGHTCOLOR
	    r->PixColors[idx] = r->PixColors[minBrightCOLOR + i];
	    r->h->pixcolor_set |= 1 << idx;
	    goto Done;
# endif
	}
	if (i >= 0 && i <= 7) {	/* normal colors */
	    r->PixColors[idx] = r->PixColors[minCOLOR + i];
	    r->h->pixcolor_set |= 1 << idx;
	    goto Done;
	}
    }
    if (!rxvt_rXParseAllocColor(r, &xcol, color))
	return;
/* XStoreColor (r->Xdisplay, r->Xcmap, XColor*); */

/*
 * FIXME: should free colors here, but no idea how to do it so instead,
 * so just keep gobbling up the colormap
 */
# if 0
    for (i = Color_Black; i <= Color_White; i++)
	if (r->PixColors[idx] == r->PixColors[i])
	    break;
    if (i > Color_White) {
	/* fprintf (stderr, "XFreeColors: r->PixColors [%d] = %lu\n", idx, r->PixColors [idx]); */
	XFreeColors(r->Xdisplay, r->Xcmap, (r->PixColors + idx), 1,
		    DisplayPlanes(r->Xdisplay, Xscreen));
    }
# endif

    r->PixColors[idx] = xcol.pixel;
    r->h->pixcolor_set |= 1 << idx;

/* XSetWindowAttributes attr; */
/* Cursor cursor; */
  Done:
    if (idx == Color_bg && !(r->Options & Opt_transparent))
	XSetWindowBackground(r->Xdisplay, r->TermWin.vt,
			     r->PixColors[Color_bg]);

/* handle Color_BD, scrollbar background, etc. */

    rxvt_set_colorfgbg(r);
    {
	XColor          fg, bg;

	fg.pixel = r->PixColors[Color_pointer];
	XQueryColor(r->Xdisplay, r->Xcmap, &fg);
	bg.pixel = r->PixColors[Color_bg];
	XQueryColor(r->Xdisplay, r->Xcmap, &bg);
	XRecolorCursor(r->Xdisplay, r->TermWin_cursor, &fg, &bg);
    }
/* the only reasonable way to enforce a clean update */
    rxvt_scr_touch(r, 0);
}

#else
# define rxvt_set_window_color(r, idx,color)	((void)0)
#endif				/* XTERM_COLOR_CHANGE */
/*----------------------------------------------------------------------*/
/*
 * find if fg/bg matches any of the normal (low-intensity) colors
 */
/* INTPROTO */
void
rxvt_set_colorfgbg(rxvt_t *r)
{
    unsigned int    i;
    const char     *xpmb = "\0";
    char            fstr[sizeof("default") + 1], bstr[sizeof("default") + 1];

    r->h->env_colorfgbg = MALLOC(sizeof("COLORFGBG=default;default;bg") + 1);
    STRCPY(fstr, "default");
    STRCPY(bstr, "default");
    for (i = Color_Black; i <= Color_White; i++)
	if (r->PixColors[Color_fg] == r->PixColors[i]) {
	    sprintf(fstr, "%d", (i - Color_Black));
	    break;
	}
    for (i = Color_Black; i <= Color_White; i++)
	if (r->PixColors[Color_bg] == r->PixColors[i]) {
	    sprintf(bstr, "%d", (i - Color_Black));
#ifdef XPM_BACKGROUND
	    xpmb = "default;";
#endif
	    break;
	}
    sprintf(r->h->env_colorfgbg, "COLORFGBG=%s;%s%s", fstr, xpmb, bstr);
    putenv(r->h->env_colorfgbg);

#ifndef NO_BRIGHTCOLOR
    r->h->colorfgbg = DEFAULT_RSTYLE;
    for (i = minCOLOR; i <= maxCOLOR; i++) {
	if (r->PixColors[Color_fg] == r->PixColors[i]
# ifndef NO_BOLD_UNDERLINE_REVERSE
	    && r->PixColors[Color_fg] == r->PixColors[Color_BD]
# endif				/* ! NO_BOLD_UNDERLINE_REVERSE */
	    /* if we wanted boldFont to have precedence */
# if 0				/* ifndef NO_BOLDFONT */
	    && r->TermWin.boldFont == NULL
# endif				/* NO_BOLDFONT */
	    )
	    r->h->colorfgbg = SET_FGCOLOR(r->h->colorfgbg, i);
	if (r->PixColors[Color_bg] == r->PixColors[i])
	    r->h->colorfgbg = SET_BGCOLOR(r->h->colorfgbg, i);
    }
#endif
}

/*----------------------------------------------------------------------*/
/*
 * Colour determination for low colour displays, routine from
 *     Hans de Goede <hans@highrise.nl>
 */

#define rSQR(x)		((float)(x)*(x))

/* EXTPROTO */
Status
rxvt_rXParseAllocColor(rxvt_t *r, XColor *screen_in_out, const char *colour)
{
    int             res = 0;

    if (!XParseColor(r->Xdisplay, r->Xcmap, colour, screen_in_out))
	rxvt_print_error("can't determine colour: %s", colour);
    else
	res = rxvt_rXAllocColor(r, screen_in_out, colour);
    return res;
}

/* EXTPROTO */
Status
rxvt_rXAllocColor(rxvt_t *r, XColor *screen_in_out, const char *colour)
{
    int             res;

    if ((res = XAllocColor(r->Xdisplay, r->Xcmap, screen_in_out)))
	return res;

    /* try again with closest match */
    if (r->Xdepth >= 4 && r->Xdepth <= 8) {
	int             i, numcol;
	int             best_pixel = 0;
	float           best_diff, diff;
	XColor         *colors;

	numcol = 0x01 << r->Xdepth;
	if ((colors = MALLOC(numcol * sizeof(XColor)))) {
	    for (i = 0; i < numcol; i++)
		colors[i].pixel = i;

	    XQueryColors(r->Xdisplay, r->Xcmap, colors, numcol);
	    best_diff = rSQR(screen_in_out->red - colors[0].red)
			+ rSQR(screen_in_out->green - colors[0].green)
			+ rSQR(screen_in_out->blue - colors[0].blue);
	    for (i = 1; i < numcol; i++) {
		diff = rSQR(screen_in_out->red - colors[i].red)
		       + rSQR(screen_in_out->green - colors[i].green)
		       + rSQR(screen_in_out->blue - colors[i].blue);
		if (diff < best_diff) {
		    best_pixel = colors[i].pixel;
		    best_diff = diff;
		}
	    }
	    *screen_in_out = colors[best_pixel];
	    FREE(colors);
	    res = XAllocColor(r->Xdisplay, r->Xcmap, screen_in_out);
	}
    }
    if (res == 0)
	rxvt_print_error("can't allocate colour: %s", colour);
    return res;
}

/*----------------------------------------------------------------------*/
/* window resizing - assuming the parent window is the correct size */
/* INTPROTO */
void
rxvt_resize_subwindows(rxvt_t *r, int width, int height)
{
    int             x = 0, y = 0;
#ifdef RXVT_GRAPHICS
    int             old_width = r->TermWin.width,
		    old_height = r->TermWin.height;
#endif

    r->TermWin.width = r->TermWin.ncol * r->TermWin.fwidth;
    r->TermWin.height = r->TermWin.nrow * r->TermWin.fheight;

    r->szHint.width = width;
    r->szHint.height = height;

/* size and placement */
    if (scrollbar_visible()) {
	rxvt_Resize_scrollBar(r);
	x = (r->scrollBar.width + 2 * r->sb_shadow);	/* vt window x coord */
	width -= x;
	if (r->Options & Opt_scrollBar_right)
	    x = 0;		/* scrollbar on right so vt window at left */
    }
    {				/* ONLYIF MENUBAR */
	if (menubar_visible()) {
	    y = menuBar_TotalHeight();	/* for placement of vt window */
	    rxvt_Resize_menuBar(r, x, 0, width, y);
	}
    }
    XMoveResizeWindow(r->Xdisplay, r->TermWin.vt, x, y, width, height + 1);

#ifdef RXVT_GRAPHICS
    if (old_width)
	rxvt_Gr_Resize(r, old_width, old_height);
#endif

    rxvt_scr_clear(r);
    rxvt_resize_pixmap(r);
    XSync(r->Xdisplay, False);
    r->h->num_scr_allow = 0;
}

/* EXTPROTO */
void
rxvt_resize_all_windows(rxvt_t *r)
{
    rxvt_szhints_recalc(r);
    XSetWMNormalHints(r->Xdisplay, r->TermWin.parent[0], &r->szHint);
    XResizeWindow(r->Xdisplay, r->TermWin.parent[0], r->szHint.width,
		  r->szHint.height);
    rxvt_resize_window(r, r->szHint.width, r->szHint.height);
}

/*
 * Redraw window after exposure or size change
 * width/height are those of the parent
 */
/* EXTPROTO */
void
rxvt_resize_window(rxvt_t *r, unsigned int width, unsigned int height)
{
    int             new_ncol, new_nrow;

    new_ncol = (width - r->szHint.base_width) / r->TermWin.fwidth;
    new_nrow = (height - r->szHint.base_height) / r->TermWin.fheight;
    if (new_ncol > MAX_COLS) {
	new_ncol = MAX_COLS;
	width = (MAX_COLS * r->TermWin.fwidth) + r->szHint.base_width;
    }
    if (new_nrow > MAX_ROWS) {
	new_nrow = MAX_ROWS;
	height = (MAX_ROWS * r->TermWin.fheight) + r->szHint.base_height;
    }
    if (r->h->old_height == -1
	|| (new_ncol != r->TermWin.ncol)
	|| (new_nrow != r->TermWin.nrow)) {
	int             curr_screen = -1;

	/* scr_reset only works on the primary screen */
	if (r->h->old_height != -1) {	/* this is not the first time thru */
	    rxvt_selection_clear(r);
	    curr_screen = rxvt_scr_change_screen(r, PRIMARY);
	}
	r->TermWin.ncol = new_ncol;
	r->TermWin.nrow = new_nrow;

	rxvt_resize_subwindows(r, width, height);
	rxvt_scr_reset(r);

	if (curr_screen >= 0)	/* this is not the first time thru */
	    rxvt_scr_change_screen(r, curr_screen);
    } else if (width != r->h->old_width || height != r->h->old_height)
	rxvt_resize_subwindows(r, width, height);
    r->h->old_width = width;
    r->h->old_height = height;
}

/*
 * Set the width/height of the window in characters.  Units are pixels.
 * good for toggling 80/132 columns
 */
/* EXTPROTO */
void
rxvt_set_widthheight(rxvt_t *r, unsigned int width, unsigned int height)
{
    XWindowAttributes wattr;

    MIN_IT(width, (MAX_COLS * r->TermWin.fwidth));
    MIN_IT(height, (MAX_ROWS * r->TermWin.fheight));
    if (width == 0 || height == 0) {
	XGetWindowAttributes(r->Xdisplay, Xroot, &wattr);
	if (width == 0)
	    width = wattr.width - r->szHint.base_width;
	if (height == 0)
	    height = wattr.height - r->szHint.base_height;
    }
    if (width != r->TermWin.width || height != r->TermWin.height) {
	width = r->szHint.base_width + width;
	height = r->szHint.base_height + height;

	XResizeWindow(r->Xdisplay, r->TermWin.parent[0], width, height);
	rxvt_resize_window(r, width, height);
#ifdef USE_XIM
	rxvt_IMSetStatusPosition(r);
#endif
    }
}

/* -------------------------------------------------------------------- *
 * -                      X INPUT METHOD ROUTINES                     - *
 * -------------------------------------------------------------------- */
#ifdef USE_XIM
/* INTPROTO */
void
rxvt_setSize(rxvt_t *r, XRectangle *size)
{
    size->x = r->TermWin.int_bwidth;
    size->y = r->TermWin.int_bwidth;
    size->width = Width2Pixel(r->TermWin.ncol);
    size->height = Height2Pixel(r->TermWin.nrow);
}

/* INTPROTO */
void
rxvt_setColor(rxvt_t *r, unsigned long *fg, unsigned long *bg)
{
    *fg = r->PixColors[Color_fg];
    *bg = r->PixColors[Color_bg];
}

/* Checking whether input method is running. */
/* INTPROTO */
Bool
rxvt_IMisRunning(rxvt_t *r)
{
    char           *p;
    Atom            atom;
    Window          win;
    char            server[IMBUFSIZ];

    /* get current locale modifier */
    if ((p = XSetLocaleModifiers(NULL)) != NULL) {
	STRCPY(server, "@server=");
	STRNCAT(server, &(p[4]), IMBUFSIZ - 9);	/* skip "@im=" */

	atom = XInternAtom(r->Xdisplay, server, False);
	win = XGetSelectionOwner(r->Xdisplay, atom);
	if (win != None)
	    return True;
    }
    return False;
}

/* EXTPROTO */
void
rxvt_IMSendSpot(rxvt_t *r)
{
    XPoint          spot;
    XVaNestedList   preedit_attr;

    if (r->h->Input_Context == NULL
	|| !r->TermWin.focus
	|| !(r->h->input_style & XIMPreeditPosition)
	|| !(r->h->event_type == KeyPress
	     || r->h->event_type == SelectionNotify
	     || r->h->event_type == ButtonRelease
	     || r->h->event_type == FocusIn)
	|| !rxvt_IMisRunning(r))
	return;

    rxvt_setPosition(r, &spot);

    preedit_attr = XVaCreateNestedList(0, XNSpotLocation, &spot, NULL);
    XSetICValues(r->h->Input_Context, XNPreeditAttributes, preedit_attr, NULL);
    XFree(preedit_attr);
}

/* EXTPROTO */
void
rxvt_setTermFontSet(rxvt_t *r, int idx)
{
    char           *string;
    long            length;
    XFontSet        prev_fontset;
    int             success = 0;

    if (idx < 0 || idx >= NFONTS)
	return;
    D_MAIN((stderr, "rxvt_setTermFontSet()"));
    prev_fontset = r->TermWin.fontset;
    r->TermWin.fontset = NULL;

    length = 0;
    if (r->h->rs[Rs_font + idx])
	length += STRLEN(r->h->rs[Rs_font + idx]) + 1;
# ifdef MULTICHAR_SET
    if (r->h->rs[Rs_mfont + idx])
	length += STRLEN(r->h->rs[Rs_mfont + idx]) + 1;
# endif
    if (length == 0 || (string = MALLOC(length + 1)) == NULL)
	r->TermWin.fontset = NULL;
    else {
	int             missing_charsetcount;
	char          **missing_charsetlist, *def_string;

	string[0] = '\0';
	if (r->h->rs[Rs_font + idx]) {
	    STRCAT(string, r->h->rs[Rs_font + idx]);
	    STRCAT(string, ",");
	}
# ifdef MULTICHAR_SET
	if (r->h->rs[Rs_mfont + idx]) {
	    STRCAT(string, r->h->rs[Rs_mfont + idx]);
	    STRCAT(string, ",");
	}
# endif
	string[STRLEN(string) - 1] = '\0';
	r->TermWin.fontset = XCreateFontSet(r->Xdisplay, string,
					    &missing_charsetlist,
					    &missing_charsetcount,
					    &def_string);
	FREE(string);
	if (r->TermWin.fontset != NULL)
	    success = 1;
    }

    if (success) {
	if (prev_fontset != NULL)
	    XFreeFontSet(r->Xdisplay, prev_fontset);
    } else
	r->TermWin.fontset = prev_fontset;
}

/* INTPROTO */
void
rxvt_setPreeditArea(rxvt_t *r, XRectangle *preedit_rect, XRectangle *status_rect, XRectangle *needed_rect)
{
    preedit_rect->x = needed_rect->width
	+ (scrollbar_visible() && !(r->Options & Opt_scrollBar_right)
	   ? (r->scrollBar.width + r->sb_shadow * 2) : 0);
    preedit_rect->y = Height2Pixel(r->TermWin.nrow - 1)
	+ ((r->menuBar.state == 1) ? menuBar_TotalHeight() : 0);

    preedit_rect->width = Width2Pixel(r->TermWin.ncol + 1) - needed_rect->width
	+ (!(r->Options & Opt_scrollBar_right)
	   ? (r->scrollBar.width + r->sb_shadow * 2) : 0);
    preedit_rect->height = Height2Pixel(1);

    status_rect->x = (scrollbar_visible()
		      && !(r->Options & Opt_scrollBar_right))
		     ? (r->scrollBar.width + r->sb_shadow * 2) : 0;
    status_rect->y = Height2Pixel(r->TermWin.nrow - 1)
	+ ((r->menuBar.state == 1) ? menuBar_TotalHeight() : 0);

    status_rect->width = needed_rect->width ? needed_rect->width
					    : Width2Pixel(r->TermWin.ncol + 1);
    status_rect->height = Height2Pixel(1);
}

/*
 * call_data has special meaning.  If it's NULL then we registered the
 * callback to rxvt_IMInstantiateCallback().  non-NULL means we did not
 */
/* ARGSUSED */
/* INTPROTO */
void
rxvt_IMDestroyCallback(XIM xim, XPointer client_data, XPointer call_data)
{
    rxvt_t         *r = _rxvt_vars;

    r->h->Input_Context = NULL;
    if (call_data == NULL)
	XRegisterIMInstantiateCallback(r->Xdisplay, NULL, NULL, NULL,
				       rxvt_IMInstantiateCallback, NULL);
}

/*
 * call_data has special meaning.  If it's NULL then we registered the
 * callback.  If it's non-NULL then we came here direct
 */
/* ARGSUSED */
/* EXTPROTO */
void
rxvt_IMInstantiateCallback(Display *display, XPointer client_data, XPointer call_data)
{
    char           *p, *s, buf[IMBUFSIZ], tmp[1024];
    char           *end, *next_s;
    XIM             xim = NULL;
    XIMStyles      *xim_styles = NULL;
    int             found;
    XPoint          spot;
    XRectangle      rect, status_rect, needed_rect;
    unsigned long   fg, bg;
    XVaNestedList   preedit_attr = NULL;
    XVaNestedList   status_attr = NULL;
    XIMCallback     ximcallback;
    rxvt_t         *r = _rxvt_vars;

    if (r->h->Input_Context)
	return;

    ximcallback.callback = rxvt_IMDestroyCallback;
    ximcallback.client_data = call_data;

    if (r->h->rs[Rs_inputMethod] && *(r->h->rs[Rs_inputMethod])) {
	STRNCPY(tmp, r->h->rs[Rs_inputMethod], sizeof(tmp) - 1);
	for (s = tmp; *s; s = next_s + 1) {
	    for (; *s && isspace(*s); s++) ;
	    if (!*s)
		break;
	    for (end = s; (*end && (*end != ',')); end++) ;
	    for (next_s = end--; ((end >= s) && isspace(*end)); end--) ;
	    *(end + 1) = '\0';

	    if (*s) {
		STRCPY(buf, "@im=");
		STRNCAT(buf, s, IMBUFSIZ - 5);
		if ((p = XSetLocaleModifiers(buf)) != NULL && *p
		    && (xim = XOpenIM(r->Xdisplay, NULL, NULL, NULL)) != NULL)
		    break;
	    }
	    if (!*next_s)
		break;
	}
    }
/* try with XMODIFIERS env. var. */
    if (xim == NULL && (p = XSetLocaleModifiers("")) != NULL && *p)
	xim = XOpenIM(r->Xdisplay, NULL, NULL, NULL);

/* try with no modifiers base */
    if (xim == NULL && (p = XSetLocaleModifiers("@im=none")) != NULL && *p)
	xim = XOpenIM(r->Xdisplay, NULL, NULL, NULL);

    if (xim == NULL)
	return;
    XSetIMValues(xim, XNDestroyCallback, &ximcallback, NULL);

    if (XGetIMValues(xim, XNQueryInputStyle, &xim_styles, NULL)
	|| !xim_styles) {
	rxvt_print_error("input method doesn't support any style");
	XCloseIM(xim);
	return;
    }
    STRNCPY(tmp, (r->h->rs[Rs_preeditType] ? r->h->rs[Rs_preeditType]
					   : "OverTheSpot,OffTheSpot,Root"),
	    sizeof(tmp) - 1);
    for (found = 0, s = tmp; *s && !found; s = next_s + 1) {
	unsigned short  i;

	for (; *s && isspace(*s); s++) ;
	if (!*s)
	    break;
	for (end = s; (*end && (*end != ',')); end++) ;
	for (next_s = end--; ((end >= s) && isspace(*end)); end--) ;
	*(end + 1) = '\0';

	if (!STRCMP(s, "OverTheSpot"))
	    r->h->input_style = (XIMPreeditPosition | XIMStatusNothing);
	else if (!STRCMP(s, "OffTheSpot"))
	    r->h->input_style = (XIMPreeditArea | XIMStatusArea);
	else if (!STRCMP(s, "Root"))
	    r->h->input_style = (XIMPreeditNothing | XIMStatusNothing);

	for (i = 0; i < xim_styles->count_styles; i++)
	    if (r->h->input_style == xim_styles->supported_styles[i]) {
		found = 1;
		break;
	    }
    }
    XFree(xim_styles);

    if (found == 0) {
	rxvt_print_error("input method doesn't support my preedit type");
	XCloseIM(xim);
	return;
    }
    if ((r->h->input_style != (XIMPreeditNothing | XIMStatusNothing))
	&& (r->h->input_style != (XIMPreeditArea | XIMStatusArea))
	&& (r->h->input_style != (XIMPreeditPosition | XIMStatusNothing))) {
	rxvt_print_error("This program does not support the preedit type");
	XCloseIM(xim);
	return;
    }
    if (r->h->input_style & XIMPreeditPosition) {
	rxvt_setSize(r, &rect);
	rxvt_setPosition(r, &spot);
	rxvt_setColor(r, &fg, &bg);

	preedit_attr = XVaCreateNestedList(0, XNArea, &rect,
					   XNSpotLocation, &spot,
					   XNForeground, fg,
					   XNBackground, bg,
					   XNFontSet, r->TermWin.fontset,
					   NULL);
    } else if (r->h->input_style & XIMPreeditArea) {
	rxvt_setColor(r, &fg, &bg);

	/*
	 * The necessary width of preedit area is unknown
	 * until create input context.
	 */
	needed_rect.width = 0;

	rxvt_setPreeditArea(r, &rect, &status_rect, &needed_rect);

	preedit_attr = XVaCreateNestedList(0, XNArea, &rect,
					   XNForeground, fg,
					   XNBackground, bg,
					   XNFontSet, r->TermWin.fontset,
					   NULL);
	status_attr = XVaCreateNestedList(0, XNArea, &status_rect,
					  XNForeground, fg,
					  XNBackground, bg,
					  XNFontSet, r->TermWin.fontset, NULL);
    }
    r->h->Input_Context = XCreateIC(xim, XNInputStyle, r->h->input_style,
				    XNClientWindow, r->TermWin.parent[0],
				    XNFocusWindow, r->TermWin.parent[0],
				    XNDestroyCallback, &ximcallback,
				    preedit_attr ? XNPreeditAttributes : NULL,
				    preedit_attr,
				    status_attr ? XNStatusAttributes : NULL,
				    status_attr, NULL);
    XFree(preedit_attr);
    XFree(status_attr);
    if (r->h->Input_Context == NULL) {
	rxvt_print_error("Failed to create input context");
	XCloseIM(xim);
    }
    if (r->h->input_style & XIMPreeditArea)
	rxvt_IMSetStatusPosition(r);
}

/* EXTPROTO */
void
rxvt_IMSetStatusPosition(rxvt_t *r)
{
    XRectangle      preedit_rect, status_rect, *needed_rect;
    XVaNestedList   preedit_attr, status_attr;

    if (r->h->Input_Context == NULL
	|| !r->TermWin.focus
	|| !(r->h->input_style & XIMPreeditArea)
	|| !rxvt_IMisRunning(r))
	return;

    /* Getting the necessary width of preedit area */
    status_attr = XVaCreateNestedList(0, XNAreaNeeded, &needed_rect, NULL);
    XGetICValues(r->h->Input_Context, XNStatusAttributes, status_attr, NULL);
    XFree(status_attr);

    rxvt_setPreeditArea(r, &preedit_rect, &status_rect, needed_rect);

    preedit_attr = XVaCreateNestedList(0, XNArea, &preedit_rect, NULL);
    status_attr = XVaCreateNestedList(0, XNArea, &status_rect, NULL);

    XSetICValues(r->h->Input_Context,
		 XNPreeditAttributes, preedit_attr,
		 XNStatusAttributes, status_attr, NULL);

    XFree(preedit_attr);
    XFree(status_attr);
}
#endif				/* USE_XIM */

/*----------------------- end-of-file (C source) -----------------------*/
